#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_site.site module

This module provides all site-related features, like Pyramid's site factory and BaseSiteRoot
which is the base implementation of any website or application.
"""

from pyramid.authorization import ALL_PERMISSIONS, Allow, Everyone
from pyramid.exceptions import NotFound
from pyramid.path import DottedNameResolver
from pyramid_zodbconn import get_connection
from zope.component import queryAdapter
from zope.component.interfaces import IPossibleSite
from zope.interface import implementer
from zope.site import LocalSiteManager
from zope.site.folder import Folder
from zope.traversing.interfaces import ITraversable

from pyams_site.interfaces import IConfigurationManager, ISiteEtcTraverser, ISiteRoot, \
    ISiteRootFactory, NewLocalSiteCreatedEvent, PYAMS_APPLICATION_DEFAULT_NAME, \
    PYAMS_APPLICATION_FACTORY_KEY, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.registry import get_current_registry, set_local_registry


__docformat__ = 'restructuredtext'


@implementer(ISiteRoot, IConfigurationManager)
class BaseSiteRoot(Folder):
    """Default site root

    A site root can be used as base application root in your ZODB.
    It's also site root responsibility to manage your local site manager.

    BaseSiteRoot defines a basic ACL which gives all permissions to system administrator,
    and 'public' permission to everyone. But this ACL is generally overriden in subclasses
    which also inherit from :py:class:`ProtectedObjectMixin
    <pyams_security.security.ProtectedObjectMixin>`.
    """

    __acl__ = [(Allow, 'system:admin', ALL_PERMISSIONS),
               (Allow, Everyone, {'public'})]

    config_klass = None


@adapter_config(name='etc', required=ISiteRoot, provides=ITraversable)
class SiteRootEtcTraverser(ContextAdapter):
    """Site root ++etc++ namespace traverser

    Gives access to local site manager from */++etc++site* URL
    """

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Traverse to site manager;
        see :py:class:`ITraversable <zope.traversing.interfaces.ITraversable>`"""
        extension = queryAdapter(self.context, ISiteEtcTraverser, name=name)
        if extension is not None:
            return extension
        raise NotFound


@adapter_config(name='site', required=ISiteRoot, provides=ISiteEtcTraverser)
def site_root_site_traverser(context):
    """Site root ++etc++site traverser extension"""
    return context.getSiteManager()


def site_factory(request):
    """Application site factory

    On application startup, this factory checks configuration to get application name and
    load it from the ZODB; if the application can't be found, configuration is scanned to
    get application factory, create a new one and create a local site manager.
    """
    conn = get_connection(request)
    root = conn.root()
    application_key = request.registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                    PYAMS_APPLICATION_DEFAULT_NAME)
    application = root.get(application_key)
    if application is None:
        factory = request.registry.settings.get(PYAMS_APPLICATION_FACTORY_KEY)
        if factory:
            resolver = DottedNameResolver()
            factory = resolver.maybe_resolve(factory)
        else:
            factory = request.registry.queryUtility(ISiteRootFactory, default=BaseSiteRoot)
        application = root[application_key] = factory()
        if IPossibleSite.providedBy(application):
            lsm = LocalSiteManager(application, default_folder=False)
            application.setSiteManager(lsm)
        try:
            # if some components require a valid and complete registry
            # with all registered utilities, they can subscribe to
            # INewLocalSiteCreatedEvent event interface
            set_local_registry(application.getSiteManager())
            get_current_registry().notify(NewLocalSiteCreatedEvent(application))
        finally:
            set_local_registry(None)
        import transaction  # pylint: disable=import-outside-toplevel
        transaction.commit()
    return application



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_site.generations module

This module provides a site generations utility to check for server timezone utility, and
zodbupdate classes renaming rules.
"""

from persistent.mapping import PersistentMapping
from zope.lifecycleevent import ObjectCreatedEvent

from pyams_site.interfaces import ISiteGenerations, SITE_GENERATIONS_KEY
from pyams_site.site import site_factory
from pyams_utils.adapter import get_annotation_adapter
from pyams_utils.factory import get_object_factory
from pyams_utils.registry import get_current_registry, get_utilities_for, query_utility, \
    set_local_registry


__docformat__ = 'restructuredtext'


def upgrade_site(request):
    """Upgrade site when needed

    This function is executed by *pyams_upgrade* console script.
    Site generations are registered named utilities providing
    :py:class:`ISiteGenerations <pyams_utils.interfaces.site.ISiteGenerations>` interface.

    Current site generations are stored into annotations for each generation adapter.
    """
    application = site_factory(request)
    if application is not None:
        try:
            set_local_registry(application.getSiteManager())
            generations = get_annotation_adapter(application, SITE_GENERATIONS_KEY,
                                                 PersistentMapping, notify=False, locate=False)
            for name, utility in sorted(get_utilities_for(ISiteGenerations),
                                        key=lambda x: x[1].order):
                if not name:
                    name = '.'.join((utility.__module__, utility.__class__.__name__))
                current = generations.get(name)
                if not current:
                    print("Upgrading {0} to generation {1}...".format(name, utility.generation))
                elif current < utility.generation:
                    print("Upgrading {0} from generation {1} to {2}...".format(name, current,
                                                                               utility.generation))
                utility.evolve(application, current)
                generations[name] = utility.generation
        finally:
            set_local_registry(None)
        import transaction  # pylint: disable=import-outside-toplevel
        transaction.commit()
    return application


def check_required_utilities(site, utilities):
    """Utility function to check for required utilities

    :param ISite site: the site manager into which configuration may be checked
    :param tuple utilities: each element of the tuple is another tuple made of the utility
        interface, the utility registration name, the utility factory and the object name when
        creating the utility, as in:

    .. code-block:: python

        REQUIRED_UTILITIES = ((ISecurityManager, '', SecurityManager, 'Security manager'),
                              (IPrincipalAnnotationUtility, '', PrincipalAnnotationUtility,
                               'User profiles'))
    """
    registry = get_current_registry()
    for interface, name, factory, default_id in utilities:
        utility = query_utility(interface, name=name)
        if utility is None:
            lsm = site.getSiteManager()
            if default_id in lsm:
                continue
            if factory is None:
                factory = get_object_factory(interface)
            utility = factory()
            registry.notify(ObjectCreatedEvent(utility))
            lsm[default_id] = utility
            lsm.registerUtility(utility, interface, name=name)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_site.generations.timezone module

This module provides timezone utility checker.
"""

from pyams_site.generations import check_required_utilities
from pyams_site.interfaces import ISiteGenerations
from pyams_utils.interfaces.timezone import IServerTimezone
from pyams_utils.registry import utility_config
from pyams_utils.timezone.utility import ServerTimezoneUtility


__docformat__ = 'restructuredtext'

REQUIRED_UTILITIES = ((IServerTimezone, '', ServerTimezoneUtility, 'Server timezone'),)


@utility_config(name='PyAMS timezone', provides=ISiteGenerations)
class TimezoneGenerationsChecker:
    """Site timezone generations checker"""

    order = 10
    generation = 1

    @staticmethod
    def evolve(site, current=None):  # pylint: disable=unused-argument
        """Check for required utilities"""
        check_required_utilities(site, REQUIRED_UTILITIES)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_site.scripts module

This module provides a "pyams_upgrade" command line utility, which can be used to
"upgrade" a site, which means creating all it's required persistent utilities and
registering them into it's local site manager.
"""

import argparse
import sys
import textwrap

from pyramid.paster import bootstrap

from pyams_site.generations import upgrade_site


__docformat__ = 'restructuredtext'


def pyams_upgrade_cmd():
    """Check for site upgrade"""
    usage = "usage: {0} config_uri".format(sys.argv[0])
    description = """Check for database upgrade.
                  Usage: pyams_upgrade production.ini
                  """
    parser = argparse.ArgumentParser(usage=usage,
                                     description=textwrap.dedent(description))
    parser.add_argument('config_uri', help='Name of configuration file')
    args = parser.parse_args()

    config_uri = args.config_uri
    env = bootstrap(config_uri)
    closer = env['closer']
    try:
        upgrade_site(env['request'])
    finally:
        closer()
